#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : napa.py
# Contents: test neurons for assembly participation
# Author  : Christian Borgelt
# History : 2014.01.31 file created
#           2014.03.26 renamed to napa.py (was cnass.py)
#           2014.05.27 modules placed in sibling directories
#-----------------------------------------------------------------------
from sys     import argv, stderr, path, version_info
from os.path import join
from time    import time
from math    import floor, ceil
from random  import seed
from bisect  import bisect
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts    import gentrain

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------
oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------
# File Reading Function
#-----------------------------------------------------------------------

def getrec (f, recseps, fldseps, blanks, comment):
    '''Read a record from a file and split it into fields.
f       file to read from
recseps record  separators
fldseps field   separators
blanks  blank   characters
comment comment characters
returns a list of the fields read or None at end of file'''
    c = f.read(1).decode()      # read the next character
    while c != '' and c in comment:
        while c != '' and c not in recseps:
            c = f.read(1).decode() # skip comment until record separator
        c = f.read(1).decode()  # consume the record separator
    if c == '': return None     # if at end of file, abort
    rec = []                    # initialize the record
    sep = fldseps +recseps      # combine all separators
    while 1:                    # field read loop
        while c != '' and c in blanks:
            c = f.read(1).decode() # skip leading blanks
        fld = ''                # initialize the field
        while c != '' and c not in sep:
            fld += c            # append character to the field
            c = f.read(1).decode()  # and read the next character
        fld = fld.strip(blanks) # remove leading/trailing blanks */
        if len(fld) > 0:        # if the field is not empty,
            rec.append(fld)     # append the field to the record
        if c == '' or c in recseps: break
        c = f.read(1).decode()  # consume field separator
    return rec                  # return the record

#-----------------------------------------------------------------------
# Statistics Functions
#-----------------------------------------------------------------------

def cpc1 (trains, width, item, alpha=1.0):
    pp = [(s,i) for i,t in trains for s in t] +[(oo,-1)]
    pp.sort()                   # convert trains to point process
    gpc = cpc = 0.0             # init. (conditional) complexity
    l = r = c = 0               # init. the window indices/counter
    while pp[l][0] < oo:        # traverse the points
        s,i = pp[l]             # get the next point/spike
        s += width;             # and the items in sliding window
        while pp[r][0] <= s:    # while still inside current window
            r += 1              # add next point to the window
        iset = set([k for s,k in pp[l:r]])
        size = len(iset)        # count sliding window per complexity
        if item in iset: x = pow(size-1,alpha); cpc += x; c += 1
        else:            x = pow(size,  alpha)
        gpc += x; l += 1        # compute and sum the statistic terms
    gpc /= float(l)             # compute average pattern complexity
    cpc /= float(c)             # (globally and conditionally)
    return (cpc-gpc) /gpc if gpc > 0 else 0.0

#-----------------------------------------------------------------------

def cpc2 (trains, width, item, alpha=1.0):
    n = len(trains)             # get the number of trains
    for i in range(n):          # find the item index
        if trains[i][0] == item: item = i
    pp = [(s,i) for i in range(n) for s in trains[i][1]] +[(oo,-1)]
    pp.sort()                   # convert trains to point process
    gpc  = cpc = 0.0            # init. (conditional) complexity
    ins  = [0 for t in trains]  # init. item containment flags
    size = l = r = c = 0        # init. the window indices/counter
    while pp[l][0] < oo:        # traverse the points
        s,i = pp[l]             # get the next point/spike
        s += width;             # and the items in sliding window
        while pp[r][0] <= s:    # while still inside current window
            j = pp[r][1]        # update items in window/item set size
            if ins[j] <= 0: size += 1
            ins[j] += 1; r += 1 # update the containment counter
        if ins[item]: x = pow(size-1,alpha); cpc += x; c += 1
        else:         x = pow(size,  alpha)
        gpc += x; l += 1        # compute and sum the statistic terms
        ins[i] -= 1             # remove the processed point/item
        if ins[i] <= 0: size -= 1  # update the containment counter
    gpc /= float(l)             # compute average pattern complexity
    cpc /= float(c)             # (globally and conditionally)
    return (cpc-gpc) /gpc if gpc > 0 else 0.0

#-----------------------------------------------------------------------

def cpc3 (trains, width, item, alpha=1.0):
    pp = [(s,i) for i,t in trains for s in t] +[(oo,-1)]
    pp.sort()                   # convert trains to point process
    gpc = cpc = 0.0             # init. (conditional) complexity
    ins = dict()                # init. item containment flags
    size = l = r = c = 0        # init. the window indices/counter
    while pp[l][0] < oo:        # traverse the points
        s,i = pp[l]             # get the next point/spike
        s += width;             # and the items in sliding window
        while pp[r][0] <= s:    # while still inside current window
            j = pp[r][1]        # update items in window/item set size
            if j in ins: ins[j] += 1
            else:        ins[j]  = 1; size += 1
            r += 1              # update the containment counter
        if item in ins: x = pow(size-1,alpha); cpc += x; c += 1
        else:           x = pow(size,  alpha)
        gpc += x; l += 1        # compute and sum the statistic terms
        ins[i] -= 1             # remove the processed point/item
        if ins[i] <= 0: del ins[i]; size -= 1
    gpc /= float(l)             # compute average pattern complexity
    cpc /= float(c)             # (globally and conditionally)
    return (cpc-gpc) /gpc if gpc > 0 else 0.0

#-----------------------------------------------------------------------
# Main Program
#-----------------------------------------------------------------------

def error (msg):
    stderr.write(msg); exit()   # print an error message and abort

#-----------------------------------------------------------------------

if __name__ == '__main__':
    desc    = 'test neurons for assembly participation'
    version = 'version 1.1 (2014.03.26)         ' \
            + '(c) 2014   Christian Borgelt'
    args    = []                # list of program arguments/options
    opts    = {'w': ['width', 0.003],  # width of window/region
               'a': ['beg',  -oo   ],  # minimum of point/time span
               'z': ['end',   oo   ],  # minimum of point/time span
               'n': ['cnt',   1000 ],  # number of surrogate data runs
               'g': ['surr',  'p'  ],  # surrogate generation method
               'd': ['rand',  'r'  ],  # random function for surrogates
               'p': ['sigma', 0.005],  # sigma for random function
               'S': ['seed',  0    ],  # seed for random numbers
               'x': ['alpha', 1.0  ],  # weighting exponent
               'l': ['train', False],  # one train per line/record
               'y': ['item',  True ],  # item in line, item first
               'r': ['recseps', '\\n' ],     # record  separators
               'f': ['fldseps', ' ,\\t' ],   # field   separators
               'b': ['blanks',  ' \\t\\r' ], # blank   characters
               'C': ['comment', '#' ],       # comment characters
               'P': ['pspfn',  ''  ] } # name of pattern spectrum file

    if len(argv) <= 1:          # if no arguments are given
        opts = dict([(o,opts[o][1]) for o in opts])
        print('usage: napa.py [options] infile [outfile]')
        print(desc); print(version)
        print('-w#      width of time window/maximum distance  '
                      +'(default: %g)' % opts['w'])
        print('-a#      start of point/time range              '
                      +'(default: auto)')
        print('-z#      end   of point/time range              '
                      +'(default: auto)')
        print('-n#      number of surrogate data sets          '
                      +'(default: %d)' % opts['n'])
        print('-g#      surrogate generation method            '
                      +'(default: %s)' % opts['g'])
        print('         r    point/spike time randomization')
        print('         d    point/spike dithering/displacement')
        print('         s    train shifting/dithering')
        print('-d#      random function type for displacements '
                      +'(default: %s)' % opts['d'])
        print('         u/r  uniform/rectangular density')
        print('         t    symmetric triangular density')
        print('         g/n  Gaussian/normal density')
        print('-p#      parameter for random function density  '
                      +'(default: %g)' % opts['p'])
        print('         (standard deviation or half base width '
                      +'of rectangle/triangle)')
        print('-S#      seed for pseudo-random numbers         '
                      +'(default: %d)' % opts['S'])
        print('-l       one train per record (item and points) '
                      +'(default: id/time pairs)')
        print('-y       no item (with -l) or item after point  '
                      +'(default: item first)')
        print('-r#      record  separators                     '
                      +'(default: "%s")' % opts['r'])
        print('-f#      field   separators                     '
                      +'(default: "%s")' % opts['f'])
        print('-b#      blank   characters                     '
                      +'(default: "%s")' % opts['b'])
        print('-C#      comment characters                     '
                      +'(default: "%s")' % opts['C'])
        print('infile   file to read trains from               '
                      +'[required]')
        print('outfile  file to write neurons statistics to    '
                      +'[optional]')
        exit()                  # print usage message and abort

    stderr.write('napa.py')     # print a startup message
    stderr.write(' - ' +desc +'\n' +version +'\n')

    # --- evaluate program options and arguments ---
    for a in argv[1:]:          # traverse the program arguments
        if a[0] != '-':         # check for a fixed argument
            if len(args) < 2: args.append(a); continue
            else: error('too many arguments\n')
        if a[1] not in opts:    # check for an option
            error('unknown option: %s\n' % a[:2])
        o = opts[a[1]]          # get the corresponding option
        v = a[2:]               # and get the option argument
        if   isinstance(o[1],True.__class__): o[1] = not o[1]
        elif isinstance(o[1], (0).__class__): o[1] = int(v)
        elif isinstance(o[1], 0.0.__class__): o[1] = float(v)
        else:                                 o[1] = v
    if len(args) < 1: error('missing arguments\n')
    opts = dict([opts[o] for o in opts])

    width   = opts['width']     # time window width/maximum distance
    beg     = opts['beg']       # beginning of allowed span of spikes
    end     = opts['end']       # end       of allowed span of spikes
    cnt     = opts['cnt']       # number of surrogate data sets
    surr    = opts['surr']      # surrogate data generation method
    rand    = opts['rand']      # random displacement function
    sigma   = opts['sigma']     # random function parameter
    rseed   = opts['seed']      # random seed (0: use time)
    alpha   = opts['alpha']     # weighting exponent
    tpr     = opts['train']     # flag for one train per line
    item    = opts['item']      # and flag for item first
    recseps = opts['recseps']   # record  separators
    fldseps = opts['fldseps']   # field   separators
    blanks  = opts['blanks']    # blank   characters
    comment = opts['comment']   # comment characters
    if width < 0: error('invalid window width %g\n'        % width)
    if sigma < 0: error('invalid sigma parameter %g\n'     % sigma)
    if cnt  <= 0: error('invalid number of data sets %d\n' % cnt)
    if surr not in 'xirdskp':   # check surrogate generation method
        error('invalid surrogate data generation %s\n' % surr)
    if rand not in 'urtgn':     # check random displacement function
        error('invalid random function type %s\n'      % rand)
    if surr == 'x': cnt = 0     # adapt number of data sets
    x = [recseps,fldseps,blanks,comment]
    if version_info[0] >= 3:    # decode ASCII escape sequences
        x = [bytes(s, 'utf-8').decode('unicode_escape') for s in x]
    else:                       # decode ASCII escape sequences
        x = [s.decode('string_escape') for  s in x]
    recseps,fldseps,blanks,comment = x

    # --- read the data set to analyze ---
    t = time()                  # start timer, print log message
    stderr.write('reading %s ... ' % args[0])
    trains = dict(); k = 1      # initialize the train dictionary
    with open(args[0], 'rb') as inp:
        while 1:                # record read loop
            rec = getrec(inp, recseps, fldseps, blanks, comment)
            if not rec: break   # read the next record
            if tpr:             # if one train per line/record
                if item: n = rec[0]; rec = rec[1:]
                else:    n = k+1# get the item/neuron identifier
                trains[n] = [x for x in [float(x) for x in line]
                             if x >= beg and x <= end]
                k += 1          # collect the trains
            else:               # if item/ point pairs
                if item: n,s = rec
                else:    s,n = rec
                s = float(s)    # collect points per item/neuron
                if s < beg or s > end: continue
                if n in trains: trains[n].append(s)
                else:           trains[n] = [s]
    trains = trains.items()     # convert to pairs (item, list of times)
    points = sum([len(s) for n,s in trains])
    stderr.write('[%d item(s), %d point(s)]' % (len(trains), points))
    stderr.write(' done [%.2fs].\n' % (time()-t))
    if beg <= -oo: beg = floor(min([min(t) for n,t in trains]))
    if end >= +oo: end = ceil (max([max(t) for n,t in trains]))
    seed(rseed if rseed != 0 else time())

    trains = sorted([(int(i),t) for i,t in trains])
    dur    = end-beg; norm = 1/float(cnt) if cnt > 0 else 0
    cpc    = cpc2
    for k in range(len(trains)):
        i,t   = trains[k]; n = len(t)
        cpcs  = []
        start = time()
        for d in range(cnt):
            t = [beg+s for s in gentrain(dur, -n)]
            s = trains[:k] +[(i,t)] +trains[k+1:]
            cpcs.append(cpc(s, width, i, alpha))
        x = cpc(trains, width, i, alpha)
        stderr.write('%3d '   % i)
        stderr.write('% .6f ' % (1.0 -bisect(sorted(cpcs), x) *norm))
        stderr.write('% .6f ' % x)
        stderr.write('[%.2fs]\n' % (time()-start))
